local M = {}

local CONST = require "render.const"

-------------------------------
-- PROJECTION FUNCTIONS
-------------------------------

-- FIXED PROJECTION
-- projection that centers content with maintained aspect ratio and optional zoom
function M.get_fixed_projection(camera, window_size, project_window_size)
	camera.zoom = camera.zoom or CONST.DEFAULT_ZOOM
	local projected_width = window_size.width / camera.zoom
	local projected_height = window_size.height / camera.zoom
	local left = -(projected_width - project_window_size.width) / 2
	local bottom = -(projected_height - project_window_size.height) / 2
	local right = left + projected_width
	local top = bottom + projected_height
	return vmath.matrix4_orthographic(left, right, bottom, top, camera.near, camera.far)
end

-- FIXED FIT PROJECTION
-- projection that centers and fits content with maintained aspect ratio
function M.get_fixed_fit_projection(camera, window_size, project_window_size)
	camera.zoom = math.min(window_size.width / project_window_size.width, window_size.height / project_window_size.height)
	return M.get_fixed_projection(camera, window_size, project_window_size)
end

-- STRETCH PROJECTION
-- projection that stretches content
function M.get_stretch_projection(camera, window_size, project_window_size)
	return vmath.matrix4_orthographic(0, project_window_size.width, 0, project_window_size.height, camera.near, camera.far)
end

-- GUI PROJECTION
function M.get_gui_projection(camera, window_size)
	return vmath.matrix4_orthographic(0, window_size.width, 0, window_size.height, camera.near, camera.far)
end

-- LIGHT PROJECTION
function M.get_light_projection(camera, window_size, project_window_size)
	return vmath.matrix4_orthographic(-10, 10, -10, 10, camera.near, camera.far)  -- Adjust based on your scene
end

-------------------------------
-- DATA STRUCTURES CONSTRUCTORS
-------------------------------

-- PREDICATES CONSTRUCTOR
function M.create_predicates(...)
	local arg = {...}
	local predicates = {}
	for _, predicate_name in pairs(arg) do
		predicates[predicate_name] = render.predicate({predicate_name})
	end
	return predicates
end

-- CAMERA CONSTRUCTOR
function M.create_camera(name, projection_fn, near, far, zoom)
	return {
		["frustum"] = {},
		["view"] = vmath.matrix4(),
		["proj"] = vmath.matrix4(),
		["proj_from_camera"] = vmath.matrix4(),
		["projection_fn"] = projection_fn,
		["near"] = near or CONST.DEFAULT_NEAR,
		["far"] = far or CONST.DEFAULT_FAR,
		["zoom"] = zoom or CONST.DEFAULT_ZOOM
	}
end

-- CLEAR COLOR CONSTRUCTOR
function M.create_clear_color()
	local clear_color = vmath.vector4(0, 0, 0, 0)
	clear_color.x = sys.get_config_number("render.clear_color_red", 0)
	clear_color.y = sys.get_config_number("render.clear_color_green", 0)
	clear_color.z = sys.get_config_number("render.clear_color_blue", 0)
	clear_color.w = sys.get_config_number("render.clear_color_alpha", 0)
	return clear_color
end

-- PARAMETERS FOR CLEARING BUFFERS CONSTRUCTOR
function M.create_clear_buffers(clear_color)
	return {
		[render.BUFFER_COLOR_BIT] = clear_color,
		[render.BUFFER_DEPTH_BIT] = 1,
		[render.BUFFER_STENCIL_BIT] = 0
	}
end

-- RENDER TARGET PARAMETERS CONSTRUCTOR
function M.create_standard_rt_color_params(format)
	return {
		format     = format or render.FORMAT_RGBA32F,
		width      = render.get_window_width(),
		height     = render.get_window_height(),
		min_filter = render.FILTER_LINEAR,
		mag_filter = render.FILTER_LINEAR,
		u_wrap     = render.WRAP_CLAMP_TO_EDGE,
		v_wrap     = render.WRAP_CLAMP_TO_EDGE
	}
end

-- RENDER TARGET PARAMETERS CONSTRUCTOR
function M.create_standard_rt_depth_params()
	return {
		format     = render.FORMAT_DEPTH,
		width      = render.get_window_width(),
		height     = render.get_window_height(),
		min_filter = render.FILTER_NEAREST,
		mag_filter = render.FILTER_NEAREST,
		u_wrap     = render.WRAP_CLAMP_TO_EDGE,
		v_wrap     = render.WRAP_CLAMP_TO_EDGE
	}
end

-- SINGLE RENDER TARGET CONSTRUCTOR
function M.create_single_render_target_data(name, format)
	assert(format ~= render.FORMAT_DEPTH)
	local parameters = {
		[render.BUFFER_COLOR0_BIT] = M.create_standard_rt_color_params(format or render.FORMAT_RGBA)
	}
	return {
		color_textures_amount = 1,
		render_target = render.render_target(name, parameters)
	}
end

-- SINGLE RENDER TARGET WITH DEPTH CONSTRUCTOR
function M.create_single_render_target_with_depth_data(name, format)
	local parameters = {
		[render.BUFFER_DEPTH_BIT] = M.create_standard_rt_depth_params()
	}
	return {
		color_textures_amount = 0,
		render_target = render.render_target(name, parameters)
	}
end

-- MULTIPLE RENDER TARGET CONSTRUCTOR
function M.create_multiple_render_target_data(name)
	local parameters = {
		[render.BUFFER_COLOR0_BIT] = M.create_standard_rt_color_params(render.FORMAT_RGBA32F),
		[render.BUFFER_COLOR1_BIT] = M.create_standard_rt_color_params(render.FORMAT_RGBA32F),
		[render.BUFFER_COLOR2_BIT] = M.create_standard_rt_color_params(render.FORMAT_RGBA),
		[render.BUFFER_DEPTH_BIT] = M.create_standard_rt_depth_params()
	}
	return {
		color_textures_amount = 3,
		render_target = render.render_target(name, parameters)
	}
end

-- PHYSICAL WINDOW SIZE CONSTRUCTOR
-- Creates structure to hold data for actual physical window dimensions.
-- Note that this value might differ from the logical dimensions that are set in the "game.project" settings.
function M.create_window()
	return {
		width = render.get_window_width(),
		height = render.get_window_height()
	}
end

-- PROJECT WINDOW SIZE CONSTRUCTOR
-- Creates structure to hold data for logical window that is set in the "game.project" dimensions.
-- Note that the actual window pixel size can change, either by device constraints or user input.
function M.create_project_window()
	return {
		width = render.get_width(),
		height = render.get_height()
	}
end

-------------------------------
-- RENDERING HELPER FUNCTIONS
-------------------------------

-- Set given render target and draw to it, then get back to default target
--
function M.to_render_target(render_target, draw_function)
	render.set_render_target(render_target)
	draw_function()
	render.set_render_target(render.RENDER_TARGET_DEFAULT)
end

function M.with_texture(binding, rt, buffer_type, draw_function)
	pprint(binding, rt, buffer_type)
	render.enable_texture(binding, rt, buffer_type)
	draw_function()
	render.disable_texture(binding)
end

function M.with_material(name, draw_function)
	render.enable_material(name)
	draw_function()
	render.disable_material()
end

function M.set_viewport_to_window(window_size)
	render.set_viewport(0, 0,  window_size.width, window_size.height)
end

-- Set given render target and draw to it, then get back to default target
--
function M.draw_to_render_target_with_material(render_target, material, draw_function)
	render.set_render_target(render_target)
	render.enable_material(material)
	draw_function()
	render.disable_material()
	render.set_render_target(render.RENDER_TARGET_DEFAULT)
end

-- Draw to a quad filling the screen used for postprocessing
-- Can draw multiple render targets and can use specific material defined in render pipeline for postprocessing
-- that uses own vertex and fragment program, where effects can be applied
--
function M.draw_to_postprocess(render_targets, textures, material, draw_function)
	render.set_view(CONST.IDENTITY_MATRIX)
	render.set_projection(CONST.IDENTITY_MATRIX)

	render.set_depth_mask(false)
	render.disable_state(render.STATE_DEPTH_TEST)
	render.disable_state(render.STATE_STENCIL_TEST)
	render.enable_state(render.STATE_BLEND)
	render.set_blend_func(render.BLEND_SRC_ALPHA, render.BLEND_ONE_MINUS_SRC_ALPHA)
	render.disable_state(render.STATE_CULL_FACE)

	render.enable_material(material)

	for i, texture in ipairs(textures) do
		print("T", i-1, texture.rt.render_target, texture.buffer_type)
		render.enable_texture(i - 1, texture.rt.render_target, texture.buffer_type)
	end

	draw_function()

	for i, texture in ipairs(textures) do
		render.disable_texture(i - 1)
	end

	render.disable_material()
end


function M.clear_all_buffers(clear_buffers)
	-- turn on depth_mask before `render.clear()` to clear it as well
	render.set_depth_mask(true)
	render.set_stencil_mask(0xff)
	render.clear(clear_buffers)
end

-- WINDOW
-- Checks if window sizes has changed (only)
function M.is_window_size_changed(window)
	return render.get_window_width() ~= window.width
	or render.get_window_height() ~= window.height
end

-- Checks if window has sizes larger than 0
function M.is_window_size_invalid()
	return (render.get_window_width() <= 0) or (render.get_window_height() <= 0)
end

return M